// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef ASTROMETRY_DEFS
#define ASTROMETRY_DEFS



#include "OmSourceParameters.h"


#include "Dal.h"



/*****************************************************
*
* Computes the equatorial rectangular coordinates
*
*****************************************************/
class EquatorialCoordinates
{
public:
    double ra;
    double dec;
    double x;       
    double y;
    double z;
    EquatorialCoordinates() : ra(0.0), dec(0.0), x(0.0), y(0.0), z(0.0) {};

    EquatorialCoordinates(double raIn, double decIn);

    EquatorialCoordinates(double xIn, double yIn, double decIn);
    EquatorialCoordinates(const EquatorialCoordinates &);


    ~EquatorialCoordinates(){};

     
    void computeEquatorialCoordinates();
    
    /**************************************************************
    *
    **************************************************************/    
    void computeSphericalCoordinates();


    double arcLength( EquatorialCoordinates & a);


};    

class data : public EquatorialCoordinates {
public:

    data() : EquatorialCoordinates(0.0, 0.0), index(0), error(0.0), significance(0.0), extension(0), obsId("") {};  
    data( int n, double a, double b, float e = 0.0, float s = 0.0, Integer1 k = 0, string obs = "") : 
      EquatorialCoordinates(a, b), index(n), error(e), significance(s), extension(k), obsId(obs)  { };

    int index;    
    float error, significance;
    Integer1 extension;
    string obsId;
};






class matchedData {
public:

    matchedData() : index(0), distance(0.0)  {};

    
    matchedData( int ind, double r) : 
      index(ind), distance(r)  {};
    int index;
    double distance;
 

};

bool compare(const matchedData &a, const matchedData & b);

bool compare1(const data &a, const data & b);

bool compare2(const OmSourceParameter &a, const OmSourceParameter & b);





#endif



