! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
 
module convert_mod
use dal
use sla
use cal

use getFilterID_mod

use omconstants_mod
public
contains


!*************************************************************************
!
!************************************************************************



!*****************************************************************
!
!
!*****************************************************************

subroutine convert1(x, y, ra0, dec0, ra, dec)
    real(double), intent(in) :: x, y, ra0, dec0
    real(double), intent(out) :: ra, dec


    call sla_dtp2s(x, y, ra0, dec0, ra, dec)

    ra = ra * radiansToDegrees
    dec = dec * radiansToDegrees


end subroutine convert1

!*****************************************************************
!
!
!*****************************************************************







!*****************************************************************
!
! Converts from RA, DEC to x and y
!
!
!*****************************************************************
subroutine convert2(ra, dec, ra0, dec0, x, y)
    real(double), intent(in) :: ra, dec, ra0, dec0
    real(double), intent(out) :: x, y
    integer :: k = 0

    k = 0;
    call sla_ds2tp(ra * degreesToRadians, dec * degreesToRadians,  &
                  ra0, dec0, x, y, k)

    if (k .ne. 0) then
        call fatal("Error calling sla_ds2tp!")

    end if
      


end subroutine convert2




!*****************************************************************
!
!
!*****************************************************************

subroutine convert3(x, y, ra0, dec0, ra, dec)

    real(double), intent(in) :: x, y, ra0, dec0
    real(double), intent(out) :: ra, dec

    call sla_dtp2s(x, y, ra0, dec0, ra, dec)

  
end subroutine convert3



!*****************************************************************
!
! Converts from RA, DEC to x and y
!
!
!*****************************************************************
subroutine convert4(ra, dec, ra0, dec0, x0, y0, plateScale, x, y)
    real(double), intent(in) :: ra, dec, ra0, dec0, x0, y0, plateScale
    real(double), intent(out) :: x, y
    integer :: k = 0
    real(double) :: dx, dy
    k = 0;
    call sla_ds2tp(ra * degreesToRadians, dec * degreesToRadians,  &
                  ra0, dec0, dx, dy, k)


    x = x0 + dx / plateScale  ! x0 is in pixels, platescale in radians 
    y = y0 - dy / plateScale  ! y0 is in pixels, platescale in radians

    if (k .ne. 0) then
        call fatal("Error calling sla_ds2tp!")

    end if
      


end subroutine convert4






!***********************************************************
!
!
!
!**********************************************************
real(double) Function getplatescale()


    integer :: filterIdInt


    !  Get the filter Id integer value
    filterIdInt = filterIdToInt("V")

    call CAL_setState(OM, filterId = filterIdInt)

 
    ! Get the Plate scale (arcsecs/pixel)
    call CAL_omGetPlateScale(0.55_double, getplateScale)
    




end function getplatescale




 
end module convert_mod





