// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef PROTOTYPES_H
#define PROTYPES_H

#include <vector>
#include <f90cxx.h>
class OMWCS;
class ImageData;




#include<iostream>











/****************************************************************
*
*
*****************************************************************/

class Alignment
{
public:
    int _iIndex, _jIndex;
    double _overlapFraction;
    double _correlation;
    int _noPixelsI, _noPixelsJ;
    bool _good;
    double _dx, _dy;
public:
    Alignment() : _iIndex(0), _jIndex(0), _overlapFraction(0), _correlation(0), 
                  _noPixelsI(0), _noPixelsJ(0), _good(true), _dx(0), _dy(0)  {};
  
    Alignment(int i, int j, double area) : 
                      _iIndex(i), _jIndex(j), _overlapFraction(area), _correlation(0), 
                      _noPixelsI(0), _noPixelsJ(0), _good(true), _dx(0), _dy(0) {};

    Alignment(int i, int j, double area, double correlation) : 
                     _iIndex(i), _jIndex(j), _overlapFraction(area), _correlation(correlation),
                     _noPixelsI(0), _noPixelsJ(0), _good(true), _dx(0), _dy(0)  {};

    Alignment(int i, int j, double area, int noIPixels, int noJPixels) : 
                     _iIndex(i), _jIndex(j), _overlapFraction(area), _correlation(0.0), 
                     _noPixelsI(noIPixels), _noPixelsJ(noJPixels), _good(true), _dx(0), _dy(0)  {};


    Alignment(const Alignment &a)
    {
      _iIndex = a._iIndex;
      _jIndex = a._jIndex;
      _overlapFraction = a._overlapFraction;
      _correlation = a._correlation;
      _noPixelsI =  a._noPixelsI; 
      _noPixelsJ =  a._noPixelsJ; 
      _good = a._good;
      _dx = a._dx;
      _dy = a._dy;
 
    }
    ~Alignment() {};



    /**************************************************************************
    *
    *
    **************************************************************************/

    bool  operator<(const Alignment &a) const 
    {
     
     
      return _overlapFraction > a._overlapFraction;
      

    };



};

/*********************************************************************
*
*
*
*********************************************************************/
//void initialise(vector<string> &fileList, vector<OMWCS> &wcsList, vector<Alignment> &alignment);





void removeTemporaryFiles();


void createFileName( string & tempFileName );

void undistortImages();






/*******************************************************************
*
*
*
*
******************************************************************/
void loadData(map<string, vector<string> > &fileListMap,
              map<string, vector<OMWCS> >  &wcsListMap);


/*********************************************************************
*
*
**********************************************************************/
void initialise(map<string, vector<string> >     &fileListMap, 
                map<string, vector<OMWCS> >      &wcsListMap, 
                map<string, vector<Alignment> >  &alignmentListMap);






void setAlignmentParameters( vector<string>       &fileList, 
                             vector<OMWCS>        &wcsList, 
                             vector<Alignment>    &alignment);





void checkAstrometry( vector<OMWCS> &wcsList, OMWCS &wc3 );


void updateImageHeaders( vector<string> &fileList, vector<OMWCS> &wcsList );


/***********************************************************************************
*
*
***********************************************************************************/


void refineImage(vector<vector<float> >  &image);
void fixPixel(vector<vector<float> > &image, int i, int j, 
	      int nx, int ny, float &tempImage);

void refineQualityMap(vector< vector<int16> > &qualityImage, vector< vector<float> > &image);


void multiplyImage( vector< vector<float> > &exposureImage, const float constant);


/***************************************************************************
*
*
*
***************************************************************************/
void alignImages(vector<Alignment> &alignmentList,
                 vector<string> &fileList,
                 vector<OMWCS>  &wcsList,
                 vector<string> &stackedFileList, 
                 vector<OMWCS>  &stackedWcsList,
                 int &noStackedFiles);





/*************************************************************
*
* Function prototypes
*************************************************************/

void setUpOffsets( int di, vector<int> &xOffset, vector<int> &yOffset);





void crossCorrelate(  OMWCS &c1, ImageData &data1,
                      OMWCS &c2,                    
                      vector<vector<float> >  &image2,
                      int di,
                      double dw,
                      vector<vector<double> >  &rImage,  
                      double &maxDx, double &maxDy, double &minR, double &maxr);




 





/************************************************************
*
*
************************************************************/
void removeBackground( ImageData &data );

void setUsablePixels( ImageData &data );

double totalCounts(vector<vector<float> > & image);


double Mean(vector<vector<float> > &);
double Mean(vector<vector<float> > &image, int il, int iu, int jl, int ju);

vector<vector<float> >  computeBackgroundImage(vector< vector<float> > &image, int dx);
float computeBackground(vector<vector<float> > &image, int nx, int ny, int i0, int j0, int dx);
void createCorrelationFile(vector<vector<double> > &image, string file1, string file2, 
                           double corr, double dx, double dy);
void updateCorrelationFile(vector<vector<double> > &image, string file1, string file2, 
                           double corr, double dx, double dy);
void loadImageData(string file, ImageData &data);
void test(  OMWCS &c);


bool isSkyImage( string file, OMWCS& );

void createOutputFile(string fileName, vector< vector<real32> > &image, 
                                       vector< vector<real32> > &exposureImage,
                                       vector< vector<int16> > &qualityImage,
                                       WcsData1 *c);

void createOutputFile(string fileName, vector< vector<real32> > &image, 
                                       vector< vector<real32> > &exposureImage,
                                       vector< vector<int16> > &qualityImage,
                                       WcsData2 *c);




void addImage(vector<vector<float> >  &stackedImage, OMWCS &stackedWcs,
              vector<vector<float> > &image, OMWCS &wcs);


void addImage(vector<vector<int16> > &stackedImage, 
              OMWCS &stackedWcs,
              vector<vector<int16> > &image, OMWCS &wcs);



void getMeanPhotometryConstants(vector<OMWCS> &stackedWcsList, OMWCS &wc3);


void  findCorrelationPeak(vector< vector<double> > &image, int di, double dw, 
			  double &maxDx, double &maxDy, double &maxR, int &numPeaks);

/*******************************************************
*
*
******************************************************/

void setPixel( float &imagePixel, double value);
	       
int getImageType( string file);

         
/*******************************************************
*
*
******************************************************/

void setPixel( int16 &imagePixel, int16 value);



/*************************************************************************************
*
* Fortran interfaces
*
************************************************************************************/
F90MODPROCPROTO( void, convert_mod, convert1)(double *, double *, double *, double*, double *, double *);
F90MODPROCPROTO( void, convert_mod, convert2)(double *, double*, double*, double*, double*, double*);
F90MODPROCPROTO( void, convert_mod, convert3)(double *, double *, double *, double*, double *, double *);


F90MODPROCPROTO(double, ominterpolate_mod, neighbourinterpolate2)(double *, double *, int *, double *, double *, double *);
F90MODPROCPROTO(void, interpolate_mod, maximizer)(int *nx, int *ny, double *r1, double *dw, double *x, double *y, double *fMax);
F90MODPROCPROTO(void, fitellipse_mod, fitellipsetest)(int *);



void process(string file1, OMWCS &, ImageData &, string file2, OMWCS &, ImageData &, 
            bool &okay, double &cr, int noStacked, double &dx, double &dy);
void loadWcsData( string file, OMWCS &wcs); 
void stackFiles(vector<string> &stackedFileList, vector<OMWCS> &stackedWcsList, 
		vector<vector<float> > &image1,    vector<vector<float> > &image2, 
                //vector<vector<int16> > &image, OMWCS &c);
		vector<vector<int16> > &image, OMWCS &c, double &deadFracMax, double &frameTimeMax);

void refineCorrelation(vector<vector<double> > &r, double dw, double &maxDx, double &maxDy, double &maxR);



#endif
