C ESA (C) 2000-2018 
C
C This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
C
C    SAS is free software: you can redistribute it and/or modify
C    it under the terms of the GNU General Public License as published by
C    the Free Software Foundation, either version 3 of the License, or
C    (at your option) any later version.
C
C    SAS is distributed in the hope that it will be useful,
C    but WITHOUT ANY WARRANTY; without even the implied warranty of
C    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C    GNU General Public License for more details.
C
C    You should have received a copy of the GNU General Public License
C    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
      program test_conic
      implicit none
      save

c..
c..tests the conic section fitting routine 
c..

c..declare
      integer          i,j

c..for generating and storing the noisy circle data
      integer          n,nmax,itype,jend
      parameter        (nmax=500)
      double precision x(nmax+3),y(nmax+3),x0,y0,angmin,angmax,dangle,
     1                 ampl,angle,rmaj,rmin,co,si,rr1,rr2,xx,yy,flength,
     2                 det,t1,t2,eccen,xfoci1,yfoci1,xfoci2,yfoci2,
     3                 aprime,bprime,cprime,dprime,eprime,fprime,
     4                 tilt,tangle,xxx(nmax+3),yyy(nmax+3),
     5                 small,large,tiny,ellcir
      parameter        (small = 1.0d-3, large=1.0d6, tiny=1.0d-12,
     1                  ellcir = 1.0d-2)


c..for the results of the fitting the noisy data
      double precision a,asig,b,bsig,c,csig,d,dsig,e,esig,f,fsig,chi


c..for the random number generator
      integer          iseed
      double precision rvec(15*nmax)
      data             iseed/-86876872/
     


c...popular formats
 01   format(1x,a,1p10e12.4)
 02   format(1x,a,1pe11.3,' +/-',1pe11.3)
 03   format(1x,'a*x**2 + b*y**2 + c*x*y + d*x + e*y + f =0 where:')
 04   format(1x,1p6e12.4)



      write(6,*) 'conic type: 1=circle 2=ellipse 3=parabola 4=hyperbola'
      read(5,*) itype

      write(6,*) 
      write(6,*) 'lets generate of a noisy data set:'


c..keep coming back to here
 05   continue
      write(6,*) 
      write(6,*) 'give number of data points (less than 500) =>'
      read(5,*), n
      n = min(n,nmax)


c..circles
      if (itype .eq. 1) then

       write(6,*) 
       write(6,*) 'give radius of circle r and its and center x0,y0 =>'
       read(5,*)  rr1,x0,y0

 06    write(6,*) 
       write(6,*) 'give the min and max angles (in degrees) to generate' 
       write(6,*) 'the noisy data between. they must be different =>'
       read(5,*)  angmin,angmax
       if (angmin .eq. angmax) goto 06

       write(6,*) 
       write(6,*) 'give the amplitude of the noise (zero = no noise) =>'
       read(5,*)  ampl

       call ranmar(iseed,rvec,2*n)

       dangle = angmax - angmin
       j = 1
       do i = 1, n
        angle = angmin + dangle*rvec(j)
        rr2   = rr1 + ampl*(1.0d0 - 2.0d0 * rvec(j+1))
        x(i)  = rr2*cosd(angle) + x0
        y(i)  = rr2*sind(angle) + y0
        j     = j + 2
       enddo





c..ellipses
      else if (itype .eq. 2) then

       write(6,*) 
       write(6,*) 'give the center of the ellipse =>'
       read(5,*)  x0,y0

       write(6,*) 
       write(6,*) 'give major,minor axis lengths =>'
       read(5,*)  t1,t2
       rmaj = max(t1,t2)
       rmin = min(t1,t2) 

       write(6,*) 
       write(6,*) 'give the acute rotation angle (-90 to 90 degrees)=>'
       read(5,*)  tilt
       if (tilt .gt. 90.0)  tilt = tilt - 180.0d0
       if (tilt .lt. -90.0) tilt = tilt + 180.0d0


 16    write(6,*) 
       write(6,*) 'give the min and max angles (in degrees) to generate' 
       write(6,*) 'the noisy data between. they must be different =>'
       read(5,*)  t1,t2
       angmin = min(t1,t2)
       angmax = max(t1,t2)
       if (angmin .eq. angmax) goto 16

       write(6,*) 
       write(6,*) 'give the amplitude of the noise (zero = no noise) =>'
       read(5,*)  ampl

       call ranmar(iseed,rvec,3*n)

       co     = cosd(tilt)
       si     = sind(tilt)
       dangle = angmax - angmin
       j = 1
       do i = 1, n
        angle = angmin + dangle*rvec(j)
        rr1   = rmaj + ampl*(1.0d0 - 2.0d0 * rvec(j+1))
        rr2   = rmin + ampl*(1.0d0 - 2.0d0 * rvec(j+2))
        xx    = rr1*cosd(angle)
        yy    = rr2*sind(angle)
        x(i)  = co*xx - si*yy + x0
        y(i)  = si*xx + co*yy + y0
        j     = j + 3
       enddo





c..parabolas
      else if (itype .eq. 3) then

c..be sure the number of points is easily divisible by 2
       n = n - mod(n,2)

       write(6,*) 
       write(6,*) 'give the vertex of the parabola =>'
       read(5,*)  x0,y0

       write(6,*) 
       write(6,*) 'give focal length =>'
       read(5,*)  flength

       write(6,*) 
       write(6,*) 'give the rotation angle (in degrees) =>'
       read(5,*)  tilt

 26    write(6,*) 
       write(6,*) 'give the min and max angles (in degrees) to generate' 
       write(6,*) 'the noisy data between. they must be different and'
       write(6,*) 'between 30.0 and 90.0 degrees =>'
       read(5,*)  t1,t2
       angmin = min(t1,t2)
       angmin = max(30.0d0,angmin)
       angmax = max(t1,t2)
       angmax = min(90.0d0,angmax)
       if (angmin .eq. angmax) goto 26

       write(6,*) 
       write(6,*) 'give the amplitude of the noise (zero = no noise) =>'
       read(5,*)  ampl

       call ranmar(iseed,rvec,4*n)

       co     = cosd(tilt)
       si     = sind(tilt)
       dangle = angmax - angmin
       j = 1
       do i=1,n/2
        angle = angmin + dangle*rvec(j)
        rr1   = flength + ampl*(1.0d0 - 2.0d0 * rvec(j+1))
        rr2   = 4.0d0*rr1*cosd(angle)/sind(angle)**2
        xx    = rr2*cosd(angle)
        yy    = rr2*sind(angle)
        x(i)  = co*xx - si*yy + x0 
        y(i)  = si*xx + co*yy + y0

c..put a different scatter on the other branch
        angle     = angmin + dangle*rvec(j+2)
        rr1       = flength + ampl*(1.0d0 - 2.0d0 * rvec(j+3))
        rr2       = 4.0d0*rr1*cosd(angle)/sind(angle)**2
        xx        =  rr2*cosd(angle)
        yy        = -rr2*sind(angle)
        x(i+n/2)  = co*xx - si*yy + x0 
        y(i+n/2)  = si*xx + co*yy + y0

        j     = j + 4
       enddo





c..hyperbolas
      else if (itype .eq. 4) then

c..be sure the number of points is easily divisible by 4
       n = n - mod(n,4)

       write(6,*) 
       write(6,*) 'give the vertex of the hyperbola =>'
       read(5,*)  x0,y0

       write(6,*) 
       write(6,*) 'give major,minor axis lengths =>'
       read(5,*)  t1,t2
       rmaj = max(t1,t2)
       rmin = min(t1,t2) 

       write(6,*) 
       write(6,*) 'give angle between the major axis and the x-axis,'
       write(6,*) 'the tilt of the hyperbola (in degrees) =>'
       read(5,*)  tilt


 36    write(6,*) 
       write(6,*) 'give the min and max t range (between -5 and 5) =>'
       read(5,*)  t1,t2
       angmin = min(t1,t2)
       angmin = max(-5.0d0,angmin)
       angmax = max(t1,t2)
       angmax = min(5.0d0,angmax)
       if (angmin .eq. angmax) goto 36

       write(6,*) 
       write(6,*) 'give the amplitude of the noise (zero = no noise) =>'
       read(5,*)  ampl

       call ranmar(iseed,rvec,12*n)

c..put a different scatter on all 4 branches of the hyperbola
       co      = cosd(tilt)
       si      = sind(tilt)
       dangle = angmax - angmin
       j = 1
       do i = 1, n/4
        angle = angmin + dangle*(1.0d0 - rvec(j))
        rr1   = rmaj + ampl*(1.0d0 - 2.0d0 * rvec(j+1))
        rr2   = rmin + ampl*(1.0d0 - 2.0d0 * rvec(j+2))
        xx    = rr1*cosh(angle) 
        yy    = rr2*sinh(angle) 
        x(i)  = co*xx - si*yy + x0
        y(i)  = si*xx + co*yy + y0

        angle = angmin + dangle*(1.0d0 - rvec(j+3))
        rr1   = rmaj + ampl*(1.0d0 - 2.0d0 * rvec(j+4))
        rr2   = rmin + ampl*(1.0d0 - 2.0d0 * rvec(j+5))
        xx    = rr1*cosh(angle) 
        yy    = -rr2*sinh(angle) 
        x(i+n/4)  = co*xx - si*yy + x0
        y(i+n/4)  = si*xx + co*yy + y0

        angle = angmin + dangle*(1.0d0 - rvec(j+6))
        rr1   = rmaj + ampl*(1.0d0 - 2.0d0 * rvec(j+7))
        rr2   = rmin + ampl*(1.0d0 - 2.0d0 * rvec(j+8))
        xx    = -rr1*cosh(angle) 
        yy    = rr2*sinh(angle) 
        x(i+n/2)  = co*xx - si*yy + x0
        y(i+n/2)  = si*xx + co*yy + y0

        angle = angmin + dangle*(1.0d0 - rvec(j+9))
        rr1   = rmaj + ampl*(1.0d0 - 2.0d0 * rvec(j+10))
        rr2   = rmin + ampl*(1.0d0 - 2.0d0 * rvec(j+11))
        xx    = -rr1*cosh(angle) 
        yy    = -rr2*sinh(angle) 
        x(i+3*n/4)  = co*xx - si*yy + x0
        y(i+3*n/4)  = si*xx + co*yy + y0

        j     = j + 12
       enddo

c..unknown type
      else
       stop 'unknown conic section type' 
      end if




c..find the best fit
      call fitconic(x,y,n,a,asig,b,bsig,c,csig,d,dsig,e,esig,f,fsig,chi)



c..report the fitting coefficients
      write(6,*) ' '
      write(6,03)
      write(6,02) 'a=',a,asig,'b=',b,bsig,'c=',c,csig,
     1            'd=',d,dsig,'e=',e,esig,'f=',f,fsig
      write(6,01) 'chi squared:',chi




c..now test what type of conic section it is 
      det = c*c - 4.0d0*a*b
      write(6,01) 'det=',det


c..either an ellipse or a circle
      if (det .lt. -small) then



c..its an ellipse
       write(6,01) 'abs(a-b)=',abs(a-b)
       write(6,01) 'abs(c)  =',abs(c)
       write(6,01) 'small   =',ellcir
       write(6,01) '1+det   =',1.0d0+det

       if (  abs(a-b) .gt. ellcir .or.
     1      (abs(a-b) .lt. ellcir .and. abs(c) .gt. ellcir) ) then
        write(6,*) ' '
        write(6,*) 'appears to be an ellipse:'


c..these formulas for the center, semimajor and semiminor axes can be found
c..in old texts on analytic geometry that discuss the general equation 
c..of the second degree. i like "plane and solid analytic geometry"
c..by charles h. astron, charles scribner's sons, new york, 1908

        x0      = (2.0d0*b*d - c*e) / det
        y0      = (2.0d0*a*e - c*d) / det
        t1      = -(a*x0**2 + b*y0**2 + c*x0*y0 + d*x0 * e*y0)
        t2      = sign(1.0d0,t1)
        tilt    = 0.5d0 * tangle(c*t2,(a-b)*t2)
        co      = cosd(tilt)
        si      = sind(tilt) 
        aprime  = a*co**2 + b*si**2 + c*co*si
        bprime  = a*si**2 + b*co**2 - c*co*si
        fprime  = (b*d*d + a*e*e - c*d*e + c*c*f - 4.0d0*a*b*f)/det
        t1      = sqrt(-fprime/aprime)
        t2      = sqrt(-fprime/bprime)
        rmaj    = max(t1,t2)
        rmin    = min(t1,t2)
        flength = sqrt(rmaj**2 - rmin**2)
        eccen   = flength/rmaj
        xfoci1 = x0 + flength*co
        yfoci1 = y0 + flength*si
        xfoci2 = x0 - flength*co
        yfoci2 = y0 - flength*si

        write(6,*)
        write(6,01) 'tilt        :',tilt
        write(6,01) 'center at   :',x0,y0
        write(6,01) 'smajor axis :',rmaj
        write(6,01) 'sminor axis :',rmin
        write(6,01) 'eccentricity:',eccen
        write(6,01) 'foci1 at    :',xfoci1,yfoci1
        write(6,01) 'foci2 at    :',xfoci2,yfoci2


c..store the fitted curve 
        xxx(1) = x0
        yyy(1) = y0
        xxx(2) = xfoci1
        yyy(2) = yfoci1
        xxx(3) = xfoci2
        yyy(3) = yfoci2
        co = cosd(tilt)
        si = sind(tilt)
        jend = 361
        do i=1,jend
         angle  = 0.0d0 + float(i-1)
         xx     = rmaj * cosd(angle) 
         yy     = rmin * sind(angle) 
         xxx(i+3) = co*xx - si*yy + x0
         yyy(i+3) = si*xx + co*yy + y0
        enddo



c..its a circle
       else
        write(6,*) ' '
        write(6,*) 'appears to be a circle:'
        x0   = -0.5d0 * d/a
        y0   = -0.5d0 * e/a
        rr1  = x0*x0 + y0*y0 - f/a
        rr2  = 0.0d0
        if (rr1 .gt. 0.0) rr2 = sqrt(rr1)
        write(6,01) 'center at  :',x0,y0
        write(6,01) 'radius r   :',rr2

c..store the fitted curve 
        xxx(1) = x0
        yyy(1) = y0
        xxx(2) = x0
        yyy(2) = y0
        xxx(3) = x0
        yyy(3) = y0
        jend = 361
        do i=1,jend
         angle   = 0.0d0 + float(i-1)
         xxx(i+3) = x0 - rr2 * cosd(angle) 
         yyy(i+3) = y0 - rr2 * sind(angle) 
        enddo
       end if





c..its a parabola
      else if (abs(det) .le. small) then
       write(6,*) ' '
       write(6,*) 'appears to be a parabola:'


c..trial value for the tilt, it may or may not be off by +/- pi
       tilt    = tangle(-c,2.0d0*b)
       co      = cosd(tilt)
       si      = sind(tilt) 
       bprime  = a*si**2 + b*co**2 - c*co*si
       dprime  = d*co + e*si
       flength = -0.25d0 * dprime/bprime

c..if the focal length is monstrous, we are off by pi/2
 100   if (flength .gt. large) then
        tilt    = tilt + 90.0d0
        co      = cosd(tilt)
        si      = sind(tilt) 
        bprime  = a*si**2 + b*co**2 - c*co*si
        dprime  = d*co + e*si
        flength = -0.25d0 * dprime/bprime
       end if


c..if the focal length is negative, we are off by pi 
       if (flength .lt. 0.0) then
        tilt    = tilt + 180.0d0
        co      = cosd(tilt)
        si      = sind(tilt) 
        bprime  = a*si**2 + b*co**2 - c*co*si
        dprime  = d*co + e*si
        flength = -0.25d0 * dprime/bprime
        goto 100
       end if
       if (tilt .gt. 360.0) tilt = tilt - 360.0

c..carry on
       eprime  = e*co - d*si
       y0      = -0.5d0 * eprime/bprime
       x0      = -(1.0d0 + eprime*y0 + bprime*y0**2)/dprime
       t1      = x0
       t2      = y0  
       x0      = t1*co - t2*si
       y0      = t1*si + t2*co
       xfoci1  = x0 + flength*co
       yfoci1  = y0 + flength*si


c..report what we got
       write(6,01) 'tilt       :',tilt
       write(6,01) 'flength    :',flength
       write(6,01) 'vertex at  :',x0,y0
       write(6,01) 'foci1 at   :',xfoci1,yfoci1


c..store the fitted curve 
       xxx(1) = x0
       yyy(1) = y0
       xxx(2) = xfoci1
       yyy(2) = yfoci1
       xxx(3) = xfoci1
       yyy(3) = yfoci1
       dangle = (90.0d0 - angmin)/float(60-1)
       jend = 2*60
       do i=1,60
        angle    = 90.0d0 - float(i-1)*dangle
        rr2      = 4.0d0*flength*cosd(angle)/sind(angle)**2
        xx       = rr2*cosd(angle)
        yy       = rr2*sind(angle)
        xxx(i+3) = co*xx - si*yy + x0 
        yyy(i+3) = si*xx + co*yy + y0

        t1          =  xx
        t2          = -yy
        xxx(i+60+3) = co*t1 - si*t2 + x0
        yyy(i+60+3) = si*t1 + co*t2 + y0
       enddo





c..its a hyperbola
      else if (det .gt. small) then
       write(6,*) ' '
       write(6,*) 'appears to be a hyperbola:'


c..these formulas for the center, semimajor and semiminor axes can be found
c..in old texts on analytic geometry that discuss the general equation 
c..of the second degree. i like "plane and solid analytic geometry"
c..by charles h. astron, charles scribner's sons, new york, 1908


       x0      = (2.0d0*b*d - c*e) / det
       y0      = (2.0d0*a*e - c*d) / det       
       t1      = -(a*x0**2 + b*y0**2 + c*x0*y0 + d*x0 * e*y0)
       t2      = sign(1.0d0,t1)
       tilt    = 0.5d0 * tangle(c*t2,(a-b)*t2)
       co      = cosd(tilt)
       si      = sind(tilt) 
       aprime  = a*co**2 + b*si**2 + c*co*si
       bprime  = a*si**2 + b*co**2 - c*co*si
       fprime  = (b*d*d + a*e*e - c*d*e + c*c*f - 4.0d0*a*b*f)/det
       t1      = sqrt(abs(fprime/aprime))
       t2      = sqrt(abs(fprime/bprime))
       rmaj    = max(t1,t2)
       rmin    = min(t1,t2)
       flength = sqrt(rmaj**2 + rmin**2)
       eccen   = flength/rmaj
       xfoci1  = x0 + flength*co
       yfoci1  = y0 + flength*si
       xfoci2  = x0 - flength*co
       yfoci2  = y0 - flength*si

       write(6,*)
       write(6,01) 'tilt        :',tilt
       write(6,01) 'center at   :',x0,y0
       write(6,01) 'smajor axis :',rmaj
       write(6,01) 'sminor axis :',rmin
       write(6,01) 'eccentricity:',eccen
       write(6,01) 'foci1 at    :',xfoci1,yfoci1
       write(6,01) 'foci2 at    :',xfoci2,yfoci2


c..store the fitted curve 
       xxx(1) = x0
       yyy(1) = y0
       xxx(2) = xfoci1
       yyy(2) = yfoci1
       xxx(3) = xfoci2
       yyy(3) = yfoci2
       co = cosd(tilt)
       si = sind(tilt)
       jend = 400
       dangle = angmax/float(100-1)
       do i=1,100
        angle    = 0.0d0 + float(i-1)*dangle
        xx       = rmaj * cosh(angle) 
        yy       = rmin * sinh(angle) 
        xxx(i+3) = co*xx - si*yy + x0
        yyy(i+3) = si*xx + co*yy + y0

        t1       =  xx
        t2       = -yy
        xxx(i+100+3) = co*t1 - si*t2 + x0
        yyy(i+100+3) = si*t1 + co*t2 + y0

        t1       =  -xx
        t2       =  yy
        xxx(i+200+3) = co*t1 - si*t2 + x0
        yyy(i+200+3) = si*t1 + co*t2 + y0

        t1       =  -xx
        t2       =  -yy
        xxx(i+300+3) = co*t1 - si*t2 + x0
        yyy(i+300+3) = si*t1 + co*t2 + y0
       enddo

      end if



c..write out the random points and the best fit for plotting
c..note the first 3 points in the curve fit give the center
c..and two foci
      open (unit=3,file='noisy_data.dat',status='unknown')
      do i = 1, n
       write(3,04) x(i), y(i)
      enddo
      close(unit=3)

      open (unit=3,file='curve.dat',status='unknown')
      do i = 1,jend + 3
       write(3,04) xxx(i), yyy(i)
      enddo
      close(unit=3)


      goto 05
      end






      double precision function tangle(numer,denom)
      implicit none
      save

c..return the angle whose tangent is numer/denom
c..declare
      double precision numer,denom,a,b,small
      parameter        (small = 1.0d-12)

      a = numer
      b = denom
      if (abs(a) .lt. small) a = 0.0d0
      tangle = atan2d(a,b)
c      write(6,115) a,b,tangle
c 115  format(1x,1p3e11.3)
      return
      end






      subroutine fitconic(x,y,n,
     1                  a,asig,b,bsig,c,csig,d,dsig,e,esig,f,fsig,chisq)
      implicit none
      save

c..this routine fits data to a general conic

c..input:
c..x(1:n) = x-coordinates of the data
c..y(1:n) = y coordinates of the data
c..n      = number of data points

c..output:
c..a,b,c,d,e,f  = general conic equation coefficient
c..asig to fsig = coefficient uncertainties, a +/- asig, b +/- bsig, and so on

c..general conic:  a * x**2 + b * y**2   + c*x*y   + d*x   + e*y   + 1 = 0


c..declare the pass
       integer          n
       double precision x(n),y(n),
     1                  a,asig,b,bsig,c,csig,d,dsig,e,esig,f,fsig,chisq

c..local variables
       integer          i,nmax,npar
       parameter        (nmax=500, npar=5)
       double precision u(nmax,npar),w(npar),v(npar,npar),bp(nmax),
     1                  ap(npar),cvm(npar,npar),wmax,thresh,sum,tol
       parameter        (tol = 1.0d-12)


c..bullet check
      if (n .gt. nmax) stop 'n > nmax in routine fitell'


c..set the u matrix and right hand side
c..the entries of u are simply the terms in front of the parameters to be fit
c..and the right hand side is everything else moved across the equal sign
      do i = 1,n
       u(i,1) = x(i)**2 
       u(i,2) = y(i)**2
       u(i,3) = x(i)*y(i)
       u(i,4) = x(i)
       u(i,5) = y(i)
       bp(i)  = -1.0d0
      enddo


c..singular value decomposition
c..a call to a double precision version of the numerical recipes routine

      call svdcmp(u,n,npar,nmax,npar,w,v)


c..massage the singular values
      wmax = 0.0d0
      do i=1,npar
       if (w(i) .gt. wmax) wmax = w(i)
      enddo
      thresh = tol * wmax
      do i=1,npar
       if (w(i) .lt. thresh) w(i) = 0.0d0
      enddo


c..backsubstitution and covariance matrix
c..return the uncertainties in the fitted parameters sqrt(cvm(i,i)) if desired
c..calls to double precision versions of the numerical recipes routines

      call svbksb(u,w,v,n,npar,nmax,npar,bp,ap)
      call svdvar(v,npar,npar,w,cvm,npar)


c..evaluate the chi-square per degree of freedom
      chisq = 0.0d0
      do i=1,n
       sum = ap(1)*x(i)**2 + ap(2)*y(i)**2 + ap(3)*x(i)*y(i) 
     1       + ap(4)*x(i) + ap(5)*y(i) 
       chisq = chisq + (bp(i) - sum)**2
      enddo
      chisq = chisq/(n - npar)


c..set the return parameters
      a    = ap(1) 
      b    = ap(2)
      c    = ap(3)
      d    = ap(4)
      e    = ap(5)
      f    = 1.0d0

c..and their uncertainties
      asig = sqrt(cvm(1,1))
      bsig = sqrt(cvm(2,2))
      csig = sqrt(cvm(3,3))
      dsig = sqrt(cvm(4,4))
      esig = sqrt(cvm(5,5))
      fsig = 0.0d0
      return
      end







      subroutine ranmar(ijkl,rvec,len) 
      implicit none 
      save

c..marsaglia and zaman random number generator. 
c..period is 2**43 with 900 million different sequences
c..the state of the generator (for restarts) is in the common block. 

c..input:
c..ijkl = seed set to negative integer to initialize or reinitialize 
c..rvec = array
c..len  = length of rvec

c..output:
c..ijkl = set to +ijkl after initialization
c..rvec = array of random numbers between 0.0 and 1.0

c..note: ijkl = 54217137 is the same as the four seeds 
c..i=12, j=34, k=56, l=78 of marsagalia and zaman 

c..declare 
      integer          ijkl,len,ivec,ij,kl,i,j,k,l,ii,jj,m,iff 
      double precision rvec(len),uni,s,t 

      integer          i97,j97 
      double precision u(97),c,cd,cm 
      common /ranset1/ u,c,cd,cm,i97,j97 

      data             iff/0/ 


c..initalization initialize even if the user forgot to set ijkl to
c..a negative value the first time.
      if (ijkl .lt. 0 .or. iff .eq. 0) then 
       iff  = 1 

       if (ijkl .lt. 0) ijkl = -ijkl 

       ij   = ijkl/30082 
       kl   = ijkl - 30082 * ij 
       i    = mod(ij/177,177) + 2 
       j    = mod(ij,177) + 2 
       k    = mod(kl/169,178) + 1 
       l    = mod(kl,169) 

       do ii=1,97 
        s = 0.0d0 
        t = 0.5d0 
        do jj =1,24 
         m = mod(mod(i*j,179)*k,179) 
         i = j 
         j = k 
         k = m 
         l = mod(53*l + 1, 169) 
         if (mod(l*m,64) .ge. 32) s = s + t 
         t = 0.5d0 * t 
        enddo
        u(ii) = s 
       enddo
       c   = 362436.0d0/16777216.0d0 
       cd  = 7654321.0d0/16777216.0d0 
       cm  = 16777213.0d0/16777216.0d0 
       i97 = 97 
       j97 = 33 
      end if 

c..normal execution starts here
      do ivec=1,len 
       uni = u(i97) - u(j97) 
       if (uni .lt. 0.0) uni = uni + 1.0d0 
       u(i97) = uni 
       i97 = i97 - 1 
       if (i97 .eq. 0) i97 = 97 
       j97 = j97 - 1 
       if (j97 .eq. 0) j97 = 97 
       c = c - cd 
       if (c .lt. 0.0) c = c + cm 
       uni = uni - c 
       if (uni .lt. 0.0) uni = uni + 1.0d0 
       rvec(ivec) = uni 
      enddo
      return 
      end 







      subroutine svdcmp(a,m,n,mp,np,w,v) 
      implicit none 
      save

c..given a matrix a, logical dimensions m by n, physical dimension mp by np, 
c..this routine computes its singular value decomposition 
c..a = u * w * vtranspose 

c..the matrix u replaces a on output. The diagonal of singular values 
c..is output as the vector w. the matrix v, not its transpose, is also output.
c..m must be greater than or equal to n; if it is smaller, then a should 
c..be filled up to a square with zero rows 


c..declare 
      integer          m,n,mp,np,i,j,k,l,its,nmax,nm,jj 
      parameter        (nmax = 500) 
      double precision a(mp,np),w(np),v(np,np),rv1(nmax),scale, 
     1                 anorm,c,f,g,h,s,x,y,z,pythag 

c..householder reduction to bidiagonal form 
      g     = 0.0d0 
      scale = 0.0d0 
      anorm = 0.0d0 
      do i=1,n    
       l      = i + 1    
       rv1(i) = scale*g   
       g      = 0.0d0 
       s      = 0.0d0    
       scale  = 0.0d0    
       if (i .le. m) then 
        do k=i,m    
         scale = scale + abs(a(k,i))  
        enddo
        if (scale .ne. 0.0) then 
         do k=i,m  
          a(k,i) = a(k,i)/scale    
          s      = s + a(k,i)*a(k,i)  
         enddo
         f      = a(i,i) 
         g      = -sign(sqrt(s),f)   
         h      = f*g - s  
         a(i,i) = f - g   

         do j=l,n    
          s = 0.0d0 
          do k=i,m  
           s = s + a(k,i)*a(k,j)  
          enddo
          f = s/h    
          do k=i,m  
           a(k,j) = a(k,j) + f*a(k,i) 
          enddo
         enddo

         do k=i,m 
          a(k,i) = scale*a(k,i)    
         enddo
        end if  
       end if    
       w(i)  = scale *g    
       g     = 0.0d0 
       s     = 0.0d0  
       scale = 0.0d0    
       if ((i .le. m) .and. (i .ne. n)) then  
        do k=l,n    
         scale = scale + abs(a(i,k))  
        enddo
        if (scale .ne. 0.0) then 
         do k=l,n  
          a(i,k) = a(i,k)/scale    
          s      = s + a(i,k)*a(i,k)  
         enddo

         f      = a(i,l) 
         g      = -sign(sqrt(s),f)   
         h      = f*g - s  
         a(i,l) = f - g   
         do k=l,n  
          rv1(k) = a(i,k)/h    
         enddo

         do j=l,m    
          s = 0.0d0 
          do k=l,n  
           s = s + a(j,k)*a(i,k)  
          enddo
          do k=l,n  
           a(j,k) = a(j,k) + s*rv1(k) 
          enddo
         enddo

         do k=l,n  
          a(i,k) = scale*a(i,k)    
         enddo
        end if  
       end if    
       anorm = max(anorm,(abs(w(i))+abs(rv1(i)))) 

c..end of householder reduction to bidiagonal form 
      enddo


c..accumulation of the right hand transformations 
      do i=n,1,-1 
       if (i .lt. n) then 

c..double division to avoid underflow 
        if (g .ne. 0.0) then 
         do j=l,n  
          v(j,i) = (a(i,j)/a(i,l))/g   
         enddo
         do j=l,n  
          s = 0.0d0 
          do k=l,n    
           s = s + a(i,k)*v(k,j)    
          enddo
          do k=l,n    
           v(k,j) = v(k,j) + s*v(k,i)   
          enddo
         enddo
        end if  

        do j=l,n    
         v(i,j) = 0.0d0 
         v(j,i) = 0.0d0 
        enddo
       end if    
       v(i,i) = 1.0d0 
       g      = rv1(i) 
       l      = i  

c..end accumulation of the right hand transformations 
      enddo


c..accumulate the left hand transforms 
      do i=min(m,n),1,-1 
       l = i+1    
       g = w(i)   
       do j=l,n    
        a(i,j) = 0.0d0 
       enddo

       if (g .ne. 0.0) then   
        g=1.0d0/g    
        do j=l,n  
         s = 0.0d0 
         do k=l,m    
          s = s + a(k,i)*a(k,j)    
         enddo
         f = (s/a(i,i))*g 
         do k=i,m    
          a(k,j) = a(k,j) + f*a(k,i)   
         enddo
        enddo
        do j=i,m    
         a(j,i) = a(j,i)*g  
        enddo

       else 
        do j= i,m   
         a(j,i) = 0.0d0 
        enddo
       end if    
       a(i,i) = a(i,i) + 1.0d0 

c..end accumulate the left hand transforms 
      enddo


c..diagonalization of the bidiagonal form 
c..loop over singular values 
      do k=n,1,-1 

c..loop over allowed iterations 
       do its=1,30   

c..test for splitting is here, note rv(1) is always zero 
        do l=k,1,-1 
         nm = l-1   
         if ((abs(rv1(l))+anorm) .eq. anorm) goto  2   
         if ((abs(w(nm))+anorm)  .eq. anorm) goto  1    
        enddo

c..cacellation of rv1(l) if l .gt. 1 
1       c = 0.0d0 
        s = 1.0d0 
        do i=l,k    
         f      = s * rv1(i)   
         rv1(i) = c * rv1(i) 
         if ((abs(f)+anorm) .eq. anorm) goto  2 
         g    = w(i) 
         h    = pythag(f,g)  
         w(i) = h 
         h    = 1.0d0/h    
         c    = (g*h)   
         s    = -(f*h)   
         do j=1,m    
          y       = a(j,nm)    
          z       = a(j,i) 
          a(j,nm) = (y*c) + (z*s)  
          a(j,i)  = -(y*s) + (z*c)  
         enddo
        enddo

c..convergence check 
2       z = w(k) 
        if (l .eq. k) then   
         if (z .lt. 0.0) then   
          w(k) = -z    
          do j=1,n    
           v(j,k) = -v(j,k)   
          enddo
         end if    
         goto  3  
        end if  
        if (its .eq. 30) stop 'no convergence in 30 iters in svdcmp' 

c..shift the bottom 2 by 2 minor 
        x  = w(l) 
        nm = k-1 
        y  = w(nm)    
        g  = rv1(nm)  
        h  = rv1(k)   
        f  = ((y-z)*(y+z)+(g-h)*(g+h))/(2.0d0*h*y)  
        g  = pythag(f,1.0d0) 
        f  = ((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x  


c..now the qr transformation 
        c = 1.0d0 
        s = 1.0d0 
        do j=l,nm   
         i      = j+1    
         g      = rv1(i) 
         y      = w(i)   
         h      = s*g    
         g      = c*g    
         z      = pythag(f,h) 
         rv1(j) = z 
         c      = f/z    
         s      = h/z    
         f      = (x*c) + (g*s)   
         g      = -(x*s) + (g*c)   
         h      = y*s    
         y      = y*c    
         do jj=1,n 
          x       = v(jj,j)  
          z       = v(jj,i)  
          v(jj,j) =  (x*c)+(z*s)   
          v(jj,i) = -(x*s)+(z*c)   
         enddo
         z    = pythag(f,h) 
         w(j) = z   

c..the rotation can be arbitrary if z is zero 
         if (z .ne. 0.0) then   
          z = 1.0d0/z    
          c = f*z  
          s = h*z  
         end if    
         f =  (c*g)+(s*y)   
         x = -(s*g)+(c*y)   
         do jj=1,m 
          y       = a(jj,j)  
          z       = a(jj,i)  
          a(jj,j) = (y*c)+(z*s)   
          a(jj,i) =-(y*s)+(z*c)   
         enddo

c..end of qr transform 
        enddo
        rv1(l) = 0.0d0 
        rv1(k) = f   
        w(k)   = x 

c..end of iteration loop
       enddo
3      continue 

c..end of loop over singular values 
      enddo
      return 
      end    







      double precision function pythag(a,b) 
      implicit none 
      save

c..computes sqrt(a**2 + b**2) without destructive underflow or overflow 

c..declare 
      double precision a,b,absa,absb 
      absa = abs(a) 
      absb = abs(b) 
      if (absa .gt. absb) then 
       pythag = absa * sqrt(1.0d0 + (absb/absa)**2) 
      else 
       if (absb .eq. 0) then 
        pythag = 0.0d0 
       else 
        pythag = absb * sqrt(1.0d0 + (absa/absb)**2) 
       end if 
      end if 
      return 
      end 








      subroutine svbksb(u,w,v,m,n,mp,np,b,x) 
      implicit none 
      save

c..solves ax=b for x when a is specified by the arrays u, w and v as returned 
c..by routine svdcmp. a is of logical dimension m by n and physical dimension 
c..mp by np. b is the input right hand side. x on output is the solution 
c..vector. no input quantities are destoyed, so one may call sequentially 
c..with different b's 

c..declare 
      integer          m,n,mp,np,nmax,i,j,jj 
      parameter        (nmax = 500) 
      double precision u(mp,np),w(np),v(np,np),b(mp),x(np),tmp(nmax),s 

c..compute utranspose * b 
      do j=1,n    
       s = 0.0d0 

c..nonzero result only if wj is nonzero; and divide 
       if (w(j) .ne. 0.0) then   
        do i=1,m    
         s = s + u(i,j)*b(i)  
        enddo
        s = s/w(j)   
       end if    
       tmp(j) = s 
      enddo


c..and matrix multiply to get the answer 
      do j=1,n    
       s = 0.0d0 
       do jj=1,n 
        s = s + v(j,jj)*tmp(jj)    
       enddo
       x(j) = s   
      enddo
      return 
      end    







      subroutine svdvar(v,ma,np,w,cvm,ncvm)  
      implicit none 
      save

c..evaluates the covariance matrix cvm of the fit for ma parameters obtained 
c..by the routine svdfit, call this routine with matrices v and w as returned 
c..by svdfit. np and ncvm are the physical dimensions of v,w and cvm. 

c..declare 
      integer          ma,np,ncvm,mmax,i,j,k 
      parameter        (mmax=20)    
      double precision v(np,np),w(np),cvm(ncvm,ncvm),wti(mmax),sum  

c..initialize 
      do i=1,ma   
       wti(i) = 0.0d0    
       if (w(i) .ne. 0.0) wti(i) = 1.0d0/(w(i)*w(i)) 
      enddo

c..sum the contributions to the covariant matrix; recipe equation 14.3.20 
      do i=1,ma   
       do j=1,i  
        sum = 0.0d0 
        do k=1,ma   
         sum = sum + v(i,k)*v(j,k)*wti(k) 
        enddo
        cvm(i,j) = sum   
        cvm(j,i) = sum   
       enddo
      enddo
      return 
      end    
