// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include "Dal.h"
#include "error.h"
#include <iostream>


#include "SasDalUtils.h"
#include "dal_utilities.h"
#include "DataSubspace.h"


#include<vector>
#include <cstring>
#include "createimage.h"
#include <cstdlib>
#include <cmath>


using namespace std;
/******************************************************
*
*******************************************************/
bool LoadVectorFloatColumnData(const Table *table, 
                      const string& columnName, 
		      int &size, vector<float> &data);

bool LoadVectorDoubleColumnData(const Table *table, 
                      const string& columnName, 
		      int &size, vector<double> &data);
void loadData(string file, vector<double> &ra, vector<double> &dec);


bool isNull( float value)
{

 
    char text[20];

    sprintf(text, "%f", value);

    if (strncmp( text, "nan", 3) == 0 || 
        strncmp( text, "NAN", 3) == 0 || 
        strncmp( text, "NaN", 3) == 0 ||
        strncmp( text, "NULL", 3) == 0 || 
        strncmp( text, "null", 3) == 0 || 
        strncmp( text, "NuLL", 3) == 0)

 
     
        return true;
    else
        return false;    


};

/***********************************************************************
*
*
*
*
*
*
***********************************************************************/
template<class T>
void compareData(vector<T> & data1, vector<T> &data2, T tolerance)
{

    for(unsigned i = 0; i < min(data1.size(), data2.size()); i++)
    {
        if( abs(data1[i] - data2[i]) > tolerance)
        {
	    cout << i << ") " << data1[i] << " " << data2[i] << " diff=" << data1[i] - data2[i] << endl;

            Sas::fatal("Wrong data values in SRCLIST table-", " ");

        }

    }

}

void checkNormalMosaicedImage();
void checkMosaicedSkyImage();
void checkMosaicedPixels(int nStack, vector<vector<float> > &mosaicedImage,  
			 vector<vector<float> > &checkImage);


/******************************************************************
*
*
*
*
*
*******************************************************************/
int main()
{

  //     checkNormalMosaicedImage();
         checkMosaicedSkyImage();


     //     Sas::fatal("Stopped for test-", " ");


}
/*********************************************************************
*
*
*
*********************************************************************/
void checkNormalMosaicedImage()
{

    /*******************************************************
    *
    * Load the mosaiced sky image
    *
    ********************************************************/

    DataSet *dataSet = dataSetServer->open( "mosaicedimage.fits", Dal::Read);

    int nx = 0, ny = 0;
    vector<vector<float> >  mosaicedImage = loadFloatImage(dataSet, "PRIMARY", nx, ny);

    vector<vector<float> >  mosaicedExposureImage = loadFloatImage(dataSet, "EXPOSURE", nx, ny);


    if(mosaicedImage.size() && mosaicedExposureImage.size() ) {};

    //    vector<vector<int16> >  loadInt16Image(const string inSet, const char * name, int &nx, int &ny);



    int nStack = dataSet->attribute("NSTACKED")->asInt();

    vector<string> fileList;

    for(int i = 1; i <= nStack; i++)
    {
        char file[100];
	//  sprintf(file, "IMAGE%03d", i);
        sprintf(file, "skyimage%02d.fits", i);

	//  fileList.push_back( dataSet->attribute(file)->asString() );
        fileList.push_back( file );
 



    }

    dataSetServer->close( dataSet ); // Finished with the file

    for(int i = 0; i < nStack; i++)
    {

        cout << "Trying to open file " << fileList[i] << endl;
        DataSet *dataSet = dataSetServer->open( fileList[i], Dal::Read);
 
        dataSetServer->close( dataSet ); // Finished with the file

    }

    vector<vector<float> >  testImage = loadFloatImage(dataSet, "PRIMARY", nx, ny);

    /***********************************************************
    *
    * Check the pixels on the mosaiced image
    *
    ************************************************************/
    checkMosaicedPixels(nStack, mosaicedImage, testImage);
    

}

/***********************************************************
*
* Check the pixels on the mosaiced image
*
************************************************************/
void checkMosaicedPixels(int nStack, vector<vector<float> > &mosaicedImage,  
			 vector<vector<float> > &checkImage)

{
    if(nStack){};

    int nx1 = (int)mosaicedImage.size();
    int ny1 = (int)mosaicedImage[0].size();
    int nx2 = (int)checkImage.size();
    int ny2 = (int)checkImage[0].size();

    if( abs(nx1 - nx2) > 5)
    {
        cout << "nx1=" << nx1 << " nx2=" << nx2 << endl;
        cout << "ny1=" << ny1 << " ny2=" << ny2 << endl;

        Sas::fatal("Wrong image dimensions-", " ");

    }

    if(abs(ny1 - ny2) > 5)
    {
        cout << "nx1=" << nx1 << " nx2=" << nx2 << endl;
        cout << "ny1=" << ny1 << " ny2=" << ny2 << endl;


        Sas::fatal("Wrong image dimensions-", " ");

    }



    double sum1 = 0.0, sum2 = 0.0;

    for(int i = 0; i < nx1; i++)
    {
        for(int j = 0; j < ny1; j++)
	{

	    if( !isNull(mosaicedImage[i][j] ) ) 
                sum1 += (double)mosaicedImage[i][j];

         


        }


    }


    for(int i = 0; i < nx2; i++)
    {
        for(int j = 0; j < ny2; j++)
	{

            if( !isNull(checkImage[i][j] ) )    
	      sum2 += (double)checkImage[i][j];


        }


    }

    //    sum2 *= (double)nStack;   

    double diff = abs(sum1 - sum2);

    if( (diff / sum1) > 0.1)
    {
      cout << "nStack=" << nStack << endl;
        cout << "nx1=" << nx1 << " ny1=" << ny1 << endl;
        cout << "nx2=" << nx2 << " ny2=" << ny2 << endl;
  
        cout << "sum1=" << sum1 << " sum2=" << sum2 << endl;
        Sas::fatal("Wrong pixel values-", " ");

    }
}


/*********************************************************************
*
*
* For checking that sky-images are stacked properly
*
*
*********************************************************************/
void checkMosaicedSkyImage()
{

    /*******************************************************
    *
    * Load the mosaiced sky image
    *
    ********************************************************/

    DataSet *dataSet = dataSetServer->open( "mosaicedskyimage.fits", Dal::Read);

    int nx = 0, ny = 0;
    vector<vector<float> >  mosaicedImage1 = loadFloatImage(dataSet, "PRIMARY", nx, ny);

    vector<vector<float> >  mosaicedExposureImage = loadFloatImage(dataSet, "EXPOSURE", nx, ny);

    int nStack = dataSet->attribute("NSTACKED")->asInt();


    /********************************************************************
    *
    * Get the list of files that have been stacked 
    *
    *
    *********************************************************************/

    vector<string> fileList;
 
    for(int i = 1; i <= nStack; i++)
    {
        char file[100];
	//   sprintf(file, "IMAGE%03d", i);
        sprintf(file, "skyimage%02d.fits", i);

	//  fileList.push_back( dataSet->attribute(file)->asString() );

        fileList.push_back( file );


    }


    dataSetServer->close( dataSet ); // Finished with the file


    for(int i = 0; i < nx; i++)
    {
        for(int j = 0; j < ny; j++)
	{
	    if( !isNull(mosaicedImage1[i][j])  && !isNull( mosaicedExposureImage[i][j]) ) 
                    mosaicedImage1[i][j] *= mosaicedExposureImage[i][j];
 
	}
    }






    /*************************************************************************
    *
    *
    *
    *************************************************************************/
    vector<vector<float> >  mosaicedImage2(nx, vector<float>(ny, 0.0));

    for(int i = 0; i < nStack; i++)
    {
        cout << "Trying to open file " << fileList[i] << endl;

        DataSet *dataSet = dataSetServer->open( fileList[i], Dal::Read);
 
        vector<vector<float> >  skyImage = loadFloatImage(dataSet, "PRIMARY", nx, ny);

        dataSetServer->close( dataSet ); // Finished with the file

        if(nx != (int)skyImage.size() || ny != (int)skyImage[0].size() )
	{
	    cout << "nx=" << nx << " ny=" << ny << endl;
            cout << "skyimage nx=" << skyImage.size() << " ny=" << skyImage[0].size(); 
            Sas::fatal("Wrong image size", " ");

        }


        for(int j = 0; j < nx; j++)
	{
            for(int k = 0; k < ny; k++)
	    {
	      

	        if( !isNull(skyImage[j][k]) )       
                    mosaicedImage2[j][k] += skyImage[j][k];
 
	    }
        }

    }

    checkMosaicedPixels( nStack,  mosaicedImage1, mosaicedImage2 );
    /*****************************************************************
    *
    *
    * 
    *
    ******************************************************************/

   

}
/***********************************************************
*
*
*
*
**********************************************************/
void loadData(string file, vector<double> &ra, vector<double> &dec)
{
    DataSet *dataSet = dataSetServer->open( file, Dal::Read);

    if (!dataSet)
    {


    }    
 
    const Table *table = dataSet->table( "SRCLIST", Dal::Read);
    if( !table) {};

    /**********************************************************************
    *
    *
    **********************************************************************/
  
    int size = 0;

 
    if ( !LoadVectorDoubleColumnData( table, "RA_CORR", size, ra))
    {

    

    }

 
    if ( !LoadVectorDoubleColumnData( table, "DEC_CORR", size, dec))
    {

    

    }

}
/********************************************************
*
*
******************************************************/
bool LoadVectorFloatColumnData(const Table *table, 
                      const string& columnName, 
                      int &size, vector<float> &data)
{
     



  
  
    size = table->rows();
   
  
  
    if (table->hasColumn(columnName))
    { 
        const ColumnData *columnData = table->column(columnName)->data();
        if (!columnData) exit(0);
	

      
        float *xData = columnData->real32Data();
       


	for (int i=0; i< size; i++) 
        {
	  data.push_back( xData[i] ); 
	 
	}
     }
         
   
    return true;
};


/********************************************************
*
*
******************************************************/
bool LoadVectorDoubleColumnData(const Table *table, 
                      const string& columnName, 
                      int &size, vector<double> &data)
{
       
    size = table->rows();
   
  
  
    if (table->hasColumn(columnName))
    { 
        const ColumnData *columnData = table->column(columnName)->data();
        if (!columnData) exit(0);
	

      
        double *xData = columnData->real64Data();
       

	for (int i=0; i< size; i++) 
        {
	    data.push_back( xData[i] ); 
	 
	}
    }
         
   
    return true;
};


